(ns puppetlabs.rbac.ldap
  (:require [clojure.set :refer [intersection rename-keys]]
            [clj-ldap.client :as clj-ldap]
            [puppetlabs.trapperkeeper.core :refer [defservice]]))

(defn connect
  "Given DS connections settings, connect to the specified DS and
   return a connection object."
  [{:keys [hostname port login password ssl]}]
  (clj-ldap/connect {:host (format "%s:%d" hostname port)
                 :ssl? (boolean ssl)
                 :bind-dn login
                 :password password}))

(defn test-connection
  "Given directory server settings data, connect to the ds and return
   any timing information. Returned is a map with two possible forms;

   if the connection is successful,
   {:connection <connection object>
    :elapsed    <time in seconds it took to connect}

   if the connection fails (either because of timeout or error):
   {:error   <either \"Connection timed out\" or the body of an exception>
    :elapsed <time in seconds until error/timeout>}"
   [settings]
   (let [timeout         (:connect_timeout settings)
         timeout         (if (number? timeout) timeout 30)
         compute-elapsed #(/ (double (- (System/nanoTime) %)) 1000000000.0)
         start           (System/nanoTime)
         connection-info (future (try {:connection (connect settings)
                                       :elapsed    (compute-elapsed start)}
                                      (catch Exception e
                                        {:error (.getMessage e)
                                         :elapsed (compute-elapsed start)})))]
     (if-let [info (deref connection-info (* 1000 timeout) nil)]
       info
       (do
         (future-cancel connection-info)
         {:error "Connection timed out"
          :elapsed (compute-elapsed start)}))))

(defn get-with [m k] (get m (keyword k)))

(def make-dn (partial format "%s=%s,%s,%s"))

(defn user-by-dn [conn settings dn]
  (let [user_lookup_attr       (keyword (settings :user_lookup_attr))
        user_display_name_attr (keyword (settings :user_display_name_attr))
        user_email_attr        (keyword (settings :user_email_attr))
        attrs                  [user_display_name_attr user_email_attr user_lookup_attr]]
    (when-let [user-info (clj-ldap/get conn dn attrs)]
      (rename-keys user-info {user_lookup_attr       :login
                              user_display_name_attr :display_name
                              user_email_attr        :email}))))

(defn group-by-dn [conn settings dn]
  (let [group_lookup_attr (keyword (settings :group_lookup_attr))
        group_name_attr   (keyword (settings :group_name_attr))
        attrs             (remove nil? [group_name_attr group_lookup_attr])]
    (when-let [group-info (clj-ldap/get conn dn attrs)]
      (rename-keys group-info {group_lookup_attr :login
                               group_name_attr   :display_name}))))

(defn user-info [{:keys [user_rdn
                             base_dn
                             user_lookup_attr] :as settings} lookup]
  (let [dn   (make-dn user_lookup_attr lookup user_rdn base_dn)
        conn (connect settings)]
    (user-by-dn conn settings dn)))

(defn group-info [{:keys [group_rdn
                          base_dn
                          group_lookup_attr] :as settings} lookup]
  (let [dn   (make-dn group_lookup_attr lookup group_rdn base_dn)
        conn (connect settings)]
    (group-by-dn conn settings dn)))

(defn groups-for-dn [conn {:keys [group_object_class
                                    group_rdn
                                    base_dn
                                    group_lookup_attr
                                    group_member_attr] :as settings} dn]
  (let [group_object_class (if (empty? group_object_class) "*" group_object_class)
        search-base        (format "%s,%s" group_rdn base_dn)
        filter             (format "(&(objectClass=%s)(%s=%s))" group_object_class group_member_attr dn)]
    (if-let [result (clj-ldap/search conn search-base {:attributes [(keyword group_lookup_attr)] :filter filter})]
      result
      [])))

(defn compute-all-groups-for-user
  "Given an LDAP connection map, the DS settings, and a DS login (the
  value for a lookup attribute like CN), return all of the groups that
  the user belongs to by walking the DS group hierarchy upwards,
  checking group membership for each ancestor group.

  Returns a sequence of group maps with :dn and a group lookup key
  (probably :cn)"
  [conn {:keys [user_rdn
                base_dn
                group_lookup_attr
                user_lookup_attr] :as settings} login]
  (let [user-dn (make-dn user_lookup_attr login user_rdn base_dn)]
    (loop [group-acc []
           next-groups (groups-for-dn conn settings user-dn)]
      (if (seq next-groups)
        (recur (concat group-acc next-groups)
               (->> next-groups
                    (mapcat (comp #(groups-for-dn conn settings %) :dn))
                    set
                    vec))
        group-acc))))

(defn filter-relevant-groups
  "Given an LDAP connection map, the DS settings, a list of relevant
  known-group-logins and a list of groups (which have :dn key), return
  a sequence of group maps found in both lists. The results are
  processed by group-by-dn."
  [conn {:keys [group_lookup_attr] :as settings} known-group-logins groups]
  (->> groups
       set
       (filter (fn [relevant-group] (some #(= (get-with relevant-group group_lookup_attr) %) known-group-logins)))
       (mapv (comp (partial group-by-dn conn settings) :dn))))

(defn groups-for-user
  "Given DS settings, a list of relevant group logins and a user's
  login value, computes and returns a flattened sequence of all of the
  user's groups filtered by the relevant group logins."
  [settings known-group-logins login]
  (if (empty? known-group-logins)
    []
    (let [conn (connect settings)]
      (->> login
           (compute-all-groups-for-user conn settings)
           (filter-relevant-groups conn settings known-group-logins)))))
