(ns puppetlabs.redrover.main
  (:gen-class)
  (:require [chic-text.tools.cli :as cli-fmt]
            [clj-yaml.core :as yaml]
            [clojure.set :refer [rename-keys] :as set]
            [clojure.string :as str]
            [clojure.tools.cli :as cli]
            [clojure.tools.logging :as log]
            [puppetlabs.trapperkeeper.config :as tk-config]
            [puppetlabs.redrover.file :refer [read-file read-json-file
                                              read-yaml-file spit-json]]
            [puppetlabs.redrover.import :as import]
            [puppetlabs.redrover.postgres :as pg]
            [puppetlabs.redrover.transform :as transform]
            [slingshot.slingshot :refer [try+ throw+]])
  (:import java.io.File))

(def default-export-output-file "./dashboard-classification-export.json")

(def default-convert-output-file "./converted-dashboard-classification.json")

(def default-db-spec {:subprotocol "postgresql"
                      :database "redrover-console"
                      :user "redrover"
                      :password "redrover"})

(def default-output-file "./dashboard-classification-export.json")

(def default-dashboard-db-config-path "/etc/puppetlabs/puppet-dashboard/database.yml")

(defn load-tk-config
  [path]
  (if path
    (try+ (tk-config/load-config path)
      (catch java.io.FileNotFoundException e
        (throw+ {:kind ::config-not-found
                 :msg (.getMessage e)})))
    {}))

(defn load-dashboard-db-spec
  [path]
  (if-not (.canRead (File. path))
    nil
    (let [db-config (read-yaml-file path)
          {:keys [host port database username password adapter]} (:production db-config)]
      {:subprotocol adapter
       :host host
       :port port
       :database database
       :user username
       :password password})))

(defn merge-cli-db-opts
  [db-spec cli-opts]
  (let [cli-opts (-> cli-opts
                   (select-keys [:db-name :db-user :db-password :db-host :db-port])
                   (rename-keys {:db-name :database, :db-user :user, :db-password :password,
                                 :db-host :host, :db-port :port}))]
    (merge db-spec
           cli-opts)))

(defn- assoc-subname
  [{:keys [host port database] :as db-spec}]
  (-> db-spec
    (assoc :subname (str (if host
                           (str "//" host
                                (if port (str ":" port))
                                "/"))
                         database))
    (dissoc :host :port :database)))

(defn config->db-spec
  "Takes an application config map and the :options map from
  cli/parse-opts, returning a complete db-spec"
  ([config] (config->db-spec config {}))
  ([config cli-options]
    (let [merged-opts (-> default-db-spec
                        (merge (get-in config [:redrover :database]))
                        (merge-cli-db-opts cli-options))]
      (-> merged-opts
        (assoc-subname)
        (update-in [:subname] #(str % (if (re-find #"\?" %) "&" "?") "characterEncoding=UTF-8"))))))

(defn export!
  [config]
  (let [{:keys [db output]} config
        groups (pg/get-dashboard-groups db)
        nodes (pg/get-dashboard-nodes db)
        exported {:groups groups, :nodes nodes}]
    (spit-json output exported)
    exported))

(def cli-options
  [["-h" "--help" "Print this help output and exit."]
   ["-d" "--database DATABASE.YML" (str "PE 3.3 dashboard DB config file. Defaults to '" default-dashboard-db-config-path "'.")]
   ["-o" "--output OUTPUT" (str "Destination file for exported and converted classification data. "
                                "The default output path for the export operation is '" default-export-output-file
                                "'; for the convert operation the default is '" default-convert-output-file "'.")]
   ["-i" "--input INPUT" (str "Source file for data to be imported or converted. "
                              "The default input path for the import operation is '" default-convert-output-file
                              "'; for the convert operation the default is '" default-export-output-file "'.")]
   ["-p" "--puppet PUPPET" "Path to the puppet executable"]
   [nil "--classifier-api-url URL" "URL to the Node Manager API"]
   [nil "--ssl-key PATH" "Path to a PEM file with the private SSL key to use to connect to the Node Manager API"]
   [nil "--ssl-cert PATH" "Path to a PEM file with the public SSL key to use to connect to the Node Manager API"]
   [nil "--ssl-ca-cert PATH" "Path to a PEM file with the CA certificate to use to connect to the Node Manager API"]
   ["-c" "--config CONFIG" "Path to HOCON configuration file(s) (not normally needed)"]
   [nil "--db-name NAME" "The name of the PE console database to connect to"]
   [nil "--db-user USER" "The username to use when connecting to the PE console database"]
   [nil "--db-password PASSWORD" "The password to use when connecting to the PE console database"]
   [nil "--db-host HOST" "The address where the PE console database can be reached"]
   [nil "--db-port PORT" "The port to use when connecting to the PE console database"]])

(def usage (str "Usage: nc_migrate command [options]"
                "\n    where `command` is either `export`, `convert`, or `import`."
                "\n\nOptions:"))

(def commands #{"export" "convert" "import"})

(defn validate-command
  "Given an argument list, returns a pair of the form
  [malformed-command? command]."
  [main-args]
  (let [command (first main-args)]
    (if (contains? commands command)
      [false command]
      [true command])))

(defn expected-error?
  [e]
  (and (map? e)
       (contains? e :kind)
       (contains? e :msg)))

(defn exit-with-msg
  ([msg] (exit-with-msg msg 1))
  ([msg code]
   (println msg)
   (System/exit code)))

(defn clean-shutdown!
  ([] (clean-shutdown! 0))
  ([code]
   (shutdown-agents)
   (System/exit code)))

(defn main*
  [& args]
  (let [{:keys [options arguments summary errors]} (cli/parse-opts args cli-options :summary-fn cli-fmt/summary)
        invalid-args? (> (count arguments) 1)
        [malformed-command? command] (validate-command arguments)
        user-error? (or errors invalid-args? malformed-command?)]
    (when errors
      (doseq [error errors] (println error)))
    (when (or (:help options) user-error?)
      (println usage)
      (println summary)
      (if user-error?
        (clean-shutdown! 2)
        (clean-shutdown! 0)))

    ;; if we've gotten here then the user didn't mess the command up or ask for help
    (case command
      "export"
      (let [{:keys [output], config-path :config} options
            database (or (:database options) default-dashboard-db-config-path)
            output (or output default-export-output-file)
            config (load-tk-config config-path)
            dashboard-db-config-path (get-in config [:redrover :console-db-config-path] database)
            _ (log/info "Loading console database configuration from"
                        (format "\"%s\"" dashboard-db-config-path))
            dashboard-db-spec (load-dashboard-db-spec dashboard-db-config-path)
            config-w-dashboard-db (update-in config [:redrover :database] merge dashboard-db-spec)
            db-spec (config->db-spec config-w-dashboard-db options)]
        (when (nil? dashboard-db-spec)
          (log/warn "The dashboard configuration file" (format "\"%s\"" dashboard-db-config-path)
                    "does not exist or is not readable, using default console database"
                    "configuration and any command-line parameters instead."))
        (export! {:db db-spec, :output output})
        (log/info "Wrote dashboard classification to" (format "\"%s\"" output)))

      "convert"
      (let [{:keys [input output]} options
            input (or input default-export-output-file)
            output (or output default-convert-output-file)
            old-data (read-json-file input)
            {:keys [bad-groups bad-nodes transformed-groups]} (transform/transform old-data)]
        (if (or (seq bad-groups) (seq bad-nodes))
          (do
            (transform/print-problems bad-groups bad-nodes)
            (println (str "Please correct the classification errors, export, and rerun this step.\n"
                          "For more information, refer to "
                          "http://docs.puppetlabs.com/pe/3.8/install_upgrade_migration_tool_conflicts.html")))
          ;; else (no bad groups or nodes)
          (do
            (spit-json output transformed-groups)
            (log/info "New classification written to" output))))

      "import"
      (let [combined-options (merge (:redrover (load-tk-config (:config options))) options)
            config (import/merge-options-with-defaults combined-options)
            input (or (:input config) default-convert-output-file)
            transformed-groups (read-json-file input)]
        (import/import! config transformed-groups)
        (log/info "Classification data imported successfully.")))

    ;; After the command is done, shutdown asynchronous thread pools & exit.
    ;; Our http-client uses thread pools indirectly, so without the
    ;; shutdown-agents call, there's a mysterious 30-40 hang on exit.
    (clean-shutdown!)))

(defn catch-and-quit
  "Calls the given thunk, catching any expected errors. If one is
  caught, prints the message and exits 1."
  [f]
  (try+ (f)
    (catch expected-error? {:keys [msg]}
      (exit-with-msg msg 1))))

(defn -main
  [& args]
  (catch-and-quit #(apply main* args)))
