(ns puppetlabs.redrover.transform
  (:require [clojure.string :as string]
            [clojure.tools.logging :as log]
            [cheshire.core :as json]
            [loom.graph :as graph]
            [loom.alg :as graph-alg]
            [slingshot.slingshot :refer [throw+]]
            [puppetlabs.classifier.storage :as storage]
            [puppetlabs.classifier.schema :refer [SerializableGroup]]
            [puppetlabs.redrover.schema :refer [DashboardExport ConvertedHierarchy]]
            [puppetlabs.redrover.hierarchy :as hierarchy]
            [puppetlabs.redrover.import :refer [blank-env-group infrastructure-group-names]]
            [schema.core :as sc]))

(defn old-groups->graph
  "Convert a sequence of old groups into a directed graph as understood by
  loom."
  [groups]
  (into {} (map (juxt :name :parents) groups)))

(defn get-old-group-ordering
  "Sort old groups topologically returning a list of group names."
  [groups]
  (-> groups
    old-groups->graph
    ;; convert to a loom directed graph
    graph/digraph
    ;; reverse edges because loom interprets them in the opposite direction
    ;; from what's in our data structure
    graph/transpose
    graph-alg/topsort))

(defn sort-old-groups
  "Sort old groups topologically so that a group will never be added
  before its parents. This handles error conditions that can occur since
  they are not well reported by the underlying graph library."
  [groups]
  (let [name->group (into {} (map (juxt :name identity) groups))
        ordering (get-old-group-ordering groups)]
    (if (nil? ordering)
      (throw+ {:kind ::malformed-groups-cycle
               :msg "The exported groups contain a cycle and cannot be converted."})
      (let [sorted-groups (map name->group ordering)]
        (if (some nil? sorted-groups)
          (throw+ {:kind ::malformed-groups-missing-parents
                   :msg "The exported groups contain references to missing groups and cannot be converted."})
          sorted-groups)))))

(defn find-group-problems
  [new-hierarchy group]
  (cond
    ;; Each test here is only valid if all previous tests pass
    (hierarchy/multiple-parents? group)
    ["Node groups can only have one parent. This node group has more than one parent."]

    (hierarchy/missing-parent? new-hierarchy group)
    [(str "The group specifies a parent that is not present in the new hierarchy.\n"
          "You may be able to resolve this error automatically by resolving errors in the parent group.")]

    (hierarchy/ancestral-nodes-mismatch? new-hierarchy group)
    [(str "Nodes that are members of a node group must also be members of all ancestor node groups.\n"
          "Some nodes in this node group are missing from one or more ancestor node groups.")]

    (contains? infrastructure-group-names (:name group))
    ["The group has the same name as a PE 3.8 infrastructure group."]))

(defn find-node-problems
  [new-hierarchy node]
  (if (and (hierarchy/has-classification? node)
           (hierarchy/node-in-group? new-hierarchy node))
    [(str "Nodes must either have classes applied directly to them or be classified through"
          " node groups, not both.\nThis node has been classified both directly and through"
          " one or more node groups.")]))

(defn print-entity-problems
  [entity-label entity problems]
  (println
    (str "Unable to migrate " entity-label ":"
         "\n" (json/encode entity {:pretty true})
         "\n" "The following problems were encountered:"
         "\n" (apply str (interpose "\n" problems))
         "\n")))

(def print-group-problems (partial print-entity-problems "group"))
(def print-node-problems (partial print-entity-problems "node"))

(defn print-problems
  [bad-groups bad-nodes]
  (doseq [[group problems] bad-groups]
    (print-group-problems group problems))
  (doseq [[node problems] bad-nodes]
    (print-node-problems node problems)))

(defn abort-retry-fail!
  []
  (println "Continue with import (skipping this group)? [Yn]")
  (case (string/lower-case (read-line))
    "y" nil
    "" nil
    "n" (throw (RuntimeException. "Import aborted by user"))
    (recur)))

(def obsolete-infrastructure-group-names
  #{"default"
    "no mcollective"
    "mcollective"
    "puppet_console"
    "puppet_master"})

(defn obsolete-group?
  [group]
  (contains? obsolete-infrastructure-group-names (:name group)))

(defn prune-node-classification
  [node]
  (let [classes-to-remove [#"^pe_puppetdb(::|$)" #"^pe_postgresql(::|$)" #"^pe_mcollective(::|$)"
                           #"^puppet_enterprise::license(::|$)"]]
    (update-in node [:classes]
               (fn [classes]
                 (into {} (for [[class params] classes]
                            (if (some #(re-find % (name class)) classes-to-remove)
                              nil
                              [class params])))))))

(defn transform-and-add-group!
  "Migrate one group into the new hierarchy. Returns a list of problems
  if transformation fails, nil if it succeeds."
  [new-hierarchy group]
  (if-let [problems (find-group-problems new-hierarchy group)]
    problems
    (hierarchy/add-old-group! new-hierarchy group)))

(defn transform-and-add-node!
  "Migrate one node into the new hierarchy. Returns a list of problems if
  transformation fails, nil if it succeeds."
  [new-hierarchy node]
  (if-let [problems (find-node-problems new-hierarchy node)]
    problems
    (hierarchy/add-old-group! new-hierarchy node)))

(defn transform-group-accumulating-problems!
  [hierarchy bad-groups group]
  (if-let [problems (transform-and-add-group! hierarchy group)]
    (assoc bad-groups group problems)
    bad-groups))

(defn transform-node-accumulating-problems!
  [hierarchy bad-nodes node]
  (if-let [problems (transform-and-add-node! hierarchy node)]
    (assoc bad-nodes node problems)
    bad-nodes))

(defn pinned-group-name
  [[first-node last-node]]
  (str "Migrated nodes " first-node "-" last-node))

(defn agent-specified-pin-groups
  "Given the old groups and old nodes from the PE 3.3 dashboard, create groups
  that set the agent-specified environment and pin all nodes (both in groups and
  listed seperately) to those groups."
  [old-groups old-nodes]
  (let [all-group-nodes (mapcat :nodes old-groups)
        all-individual-nodes (mapcat :nodes old-nodes)
        pinned-nodes (->> (concat all-group-nodes all-individual-nodes)
                          (apply sorted-set)
                          seq)
        partitions (partition-all 100 pinned-nodes)]
    (for [nodes partitions
          :let [group-name (if (> (count partitions) 1)
                             (pinned-group-name ((juxt first last) nodes))
                             "Migrated nodes")]]
      (assoc (blank-env-group group-name)
             :description "Nodes migrated from PE 3.3 set their own environment"
             :rule (hierarchy/nodes->pinned-nodes-rule nodes)))))

(def JsonHierarchy
  "The immediate result of reading a converted classification as JSON."
  [(assoc SerializableGroup
          :id sc/Str
          :parent sc/Str)])

(defn validate-dashboard-export-or-throw
  [dashboard-export]
  (when-let [validation-failures (sc/check DashboardExport dashboard-export)]
    (log/debug (str "The input to the convert command caused the following schema failure: "
                    validation-failures))
    (if (nil? (sc/check JsonHierarchy dashboard-export))
      (throw+ {:kind ::invalid-input-file
               :msg (str "The input file has already been converted,"
                         " so it cannot be converted again.\nPlease re-run"
                         " this step using the output of the \"export\""
                         " command as the input.")})
      (throw+ {:kind ::invalid-input-file
               :msg (str "The input file contains valid JSON,"
                         " but it does not appear to represent a valid PE 3.3"
                         " dashboard classification.\nPlease re-run this step"
                         " using the output of the \"export\" command as the input.")}))))


(defn transform
  "Given a list of groups and nodes from a 3.3 classification transform
  them to a hierarchy for a 3.7+ classification. Groups that do not map
  cleanly to the new model will not be imported. The new hierarchy will
  be a list of plain clojure data structures ready to be serialized and
  imported into a 3.7+ classifier. Groups that do not get imported will
  be listed along with the reasons they were not imported."
  [dashboard-export]
  (validate-dashboard-export-or-throw dashboard-export)

  (let [{old-groups :groups, old-nodes :nodes} dashboard-export
        new-hierarchy (hierarchy/new-hierarchy)
        sorted-old-groups (sort-old-groups old-groups)
        bad-groups (reduce (partial transform-group-accumulating-problems! new-hierarchy)
                           {} (remove obsolete-group? sorted-old-groups))
        bad-nodes (reduce (partial transform-node-accumulating-problems! new-hierarchy)
                          {} (->> old-nodes
                               (filter hierarchy/has-classification?)
                               (map prune-node-classification)))]
    (doseq [agent-specified-group (agent-specified-pin-groups old-groups old-nodes)]
      (storage/create-group new-hierarchy agent-specified-group))
    {:bad-groups bad-groups
     :bad-nodes bad-nodes
     :transformed-groups (storage/get-groups new-hierarchy)}))
